#!/usr/bin/env python

# This code is part of the 'enjuewemela' game
# License: GPLv3
# Main author: Facundo Batista
# Code, bug tracker, etc:
#   https://launchpad.net/enjuewemela/
#
"""Main program for the game."""

from __future__ import division

import glob
import logging
import os
import random
import time

from xdg.BaseDirectory import xdg_config_home

from config_schema import ConfigSchema
try:
    from configglue.parser import SchemaConfigParser
except ImportError:
    # old way to do it
    from configglue.pyschema.parser import SchemaConfigParser

import cocos
from cocos.director import director
from cocos import actions, scene
from cocos.text import Label

import pyglet

from pyglet.window import key as keyboard

import logic_board
import stats
import remove_effects

from gamemenu import MainMenu, set_options
from gems import gems
from config import (
    ARISTA,
    BASEDIR,
    BASE_COLOR,
    BORDER,
    BUBBLE_LIFE,
    FALL_TIME,
    PIECE_TRIP_TIME,
)
from sound import sounds


class RepeatedMatcher(object):
    """Shows effects on repeated matching."""

    # how much time the "hit" stays supervised
    LAPSE = 3

    def __init__(self, layer):
        self.simult_hits = 0
        self.layer = layer

        # which sound to play and message to show when N simultaneous hits
        # (put this here because it needs '_' to be alive)
        self.EVENT = {
            3: ("cheer_good", _("Good!")),
            4: ("cheer_excelent", _("Excelent!")),
            5: ("cheer_awesome", _("Awesome!")),
        }

        self._limit_low = min(self.EVENT)
        self._limit_high = max(self.EVENT)
        self._exceeding = self.EVENT[self._limit_high]

        win_w, win_h = director.get_window_size()
        self.center = (win_w // 2, win_h // 2)

    def hit(self, quantity):
        """One or more matches happened."""
        self.simult_hits += quantity
        pyglet.clock.schedule_once(self._ended, self.LAPSE, quantity)

        # check if we need to play a sound
        if self.simult_hits < self._limit_low:
            return

        # yes, choose one, or use the exceeding
        if self.simult_hits > self._limit_high:
            to_play, to_show = self._exceeding
        else:
            to_play, to_show = self.EVENT[self.simult_hits]
        sounds.play(to_play)
        self._bubble(to_show)

    def _ended(self, _, quantity):
        """Ended the time lapse for a hit."""
        self.simult_hits -= quantity

    def _bubble(self, message):
        """Show the message."""
        lab = Label(message, self.center, font_name="Komika Krak",
                    font_size=50, color=(255, 0, 55, 255), anchor_x="center")
        self.layer.add(lab)

        def remove(_):
            """Removes the label."""
            self.layer.remove(lab)

        lab.do(actions.FadeOut(BUBBLE_LIFE))
        lab.do(actions.MoveBy((0, 70), BUBBLE_LIFE))
        pyglet.clock.schedule_once(remove, BUBBLE_LIFE)


class ConfigManager(object):
    """Keeps the user configuration."""

    def __init__(self):
        self.parser = SchemaConfigParser(ConfigSchema())
        self.fname = os.path.join(xdg_config_home, 'enjuewemela.conf')
        self.parser.read(self.fname)

    def __getitem__(self, key):
        """Dynamically sets and gets values."""
        return self.parser.get('__main__', key)

    def __setitem__(self, key, value):
        """Dynamically sets and gets values."""
        self.parser.set('__main__', key, value)
        self.parser.save(self.fname)


class BackgroundLayer(cocos.layer.Layer):
    """Background layer for all the game."""

    def __init__(self, fname):
        super(BackgroundLayer, self).__init__()
        self._image = sp = cocos.sprite.Sprite(fname)
        win_w, win_h = director.get_window_size()
        sp.scale = win_w / sp.width
        sp.position = win_w // 2, win_h // 2
        self.add(sp)

    def _set_opacity(self, val):
        """Set opacity to the image."""
        self._image.opacity = val

    opacity = property(fset=_set_opacity)


class Board(cocos.draw.Canvas):
    """Static layer for the board."""

    def __init__(self):
        cocos.draw.Canvas.__init__(self)

        win_w, win_h = director.get_window_size()
        self.ini_x = (win_w - ARISTA * 8) // 2
        self.ini_y = (win_h - ARISTA * 8) // 2
        self.center = (win_w // 2, win_h // 2)

    def render(self):
        """Render the board."""
        ini_x, ini_y = self.ini_x, self.ini_y
        self.set_endcap(cocos.draw.BUTT_CAP)

        # big white base
        self.set_stroke_width(ARISTA * 8)
        self.set_color((255, 255, 255, 140))
        self.move_to((ini_x + ARISTA * 4, ini_y))
        self.line_to((ini_x + ARISTA * 4, ini_y + ARISTA * 8))

        # two colors for the slots
        color1 = BASE_COLOR + (200,)
        color2 = BASE_COLOR + (100,)

        # slots
        self.set_stroke_width(ARISTA)
        base_x = ini_x + ARISTA // 2   # vertical paint strokes
        for x in range(8):
            for y in range(8):
                if (-1) ** (x + y) == -1:
                    self.set_color(color1)
                else:
                    self.set_color(color2)

                pos_move_to = (base_x + ARISTA * x, ini_y + ARISTA * y)
                self.move_to(pos_move_to)

                pos_line_to = (base_x + ARISTA * x, ini_y + ARISTA * (y + 1))
                self.line_to(pos_line_to)

        # border!
        self.set_stroke_width(BORDER)
        board = ARISTA * 8
        self.set_color(BASE_COLOR + (255,))

        # border, bottom line
        self.move_to((ini_x - BORDER,         ini_y - BORDER // 2))
        self.line_to((ini_x + BORDER + board, ini_y - BORDER // 2))

        # border, left line
        self.move_to((ini_x - BORDER // 2, ini_y - BORDER))
        self.line_to((ini_x - BORDER // 2, ini_y + BORDER + board))

        # border, right line
        self.move_to((ini_x + board + BORDER // 2, ini_y - BORDER))
        self.line_to((ini_x + board + BORDER // 2, ini_y + BORDER + board))

        # border, top line
        self.move_to((ini_x - BORDER,         ini_y + board + BORDER // 2))
        self.line_to((ini_x + BORDER + board, ini_y + board + BORDER // 2))

    def get_slot_center(self, pos_x, pos_y):
        """Return the coordinates of the center of the given slot."""
        x = self.ini_x + ARISTA * pos_x + ARISTA // 2
        y = self.ini_y + ARISTA * pos_y + ARISTA // 2
        return x, y

    def get_slot_from_cords(self, coord_x, coord_y):
        """Return the slot for given coordinates."""
        bx = coord_x - self.ini_x
        pos_x = bx // ARISTA
        if not (0 <= pos_x < 8):
            return None

        by = coord_y - self.ini_y
        pos_y = by // ARISTA
        if not (0 <= pos_y < 8):
            return None

        return int(pos_x), int(pos_y)

    def get_cords_from_slot(self, pos_x, pos_y):
        """Return the coordinates for given slot."""
        x = self.ini_x + pos_x * ARISTA
        y = self.ini_y + pos_y * ARISTA
        return x, y


class PieceHolder(object):
    """Class to handle pieces."""

    def __init__(self):
        self.slot2piece = {}
        self.piece2slot = {}
        self.logger = logging.getLogger('PieceHolder')

    def __str__(self):
        """Shows itself as a board."""
        return str(self.get_board())

    def set_piece(self, piece, slot_x, slot_y):
        """Set a piece in a slot."""
        self.logger.debug("set piece: %s in (%s, %s)", piece, slot_x, slot_y)
        self.piece2slot[piece] = slot_x, slot_y
        self.slot2piece[slot_x, slot_y] = piece

    def remove_piece(self, slot_x, slot_y):
        """Remove a piece from a slot."""
        piece = self.slot2piece.pop((slot_x, slot_y))
        self.logger.debug("remove piece from (%s, %s): %s",
                          slot_x, slot_y, piece)
        del self.piece2slot[piece]
        return piece

    def get_piece(self, slot_x, slot_y):
        """Return the piece in the slot, None if any."""
        try:
            return self.slot2piece[slot_x, slot_y]
        except KeyError:
            return None

    def get_slot(self, piece):
        """Return the slot where that piece is."""
        return self.piece2slot[piece]

    def get_board(self):
        """Return the complete board."""
        board = []
        for y in range(8):
            row = []
            for x in range(8):
                try:
                    piece = self.slot2piece[x, 7 - y]
                except KeyError:
                    row.append(None)
                else:
                    row.append(piece.pos_id)
            board.append(row)
        return logic_board.FakeBoard(board)

    def check_matches(self):
        """Tell in which positions there're matches."""
        return logic_board.detect(self.get_board())

    def future_guess(self):
        """Guess which piece the user could move to have a match."""
        return logic_board.future_guess(self.get_board())

    def piece_fall(self, _, piece, steps):
        """Make the piece fall."""
        # fix position as we may be called just a little before the piece
        # finished moving
        piece.position = piece.orig_pos

        # piece to slot
        slot_x, old_slot_y = self.piece2slot[piece]
        new_slot_y = old_slot_y - steps
        self.piece2slot[piece] = slot_x, new_slot_y

        # slot to piece
        del self.slot2piece[slot_x, old_slot_y]
        self.slot2piece[slot_x, new_slot_y] = piece

    def get_all_similar(self, piece):
        """Return all the slots that has a piece like that."""
        res = [(p, s) for p, s in self.piece2slot.iteritems()
               if p.name == piece.name]
        return res

    def try_swap(self, pc1, pc2):
        """Try to swap the pieces."""
        # get slots
        sl1 = self.piece2slot[pc1]
        sl2 = self.piece2slot[pc2]
        # swap and test
        self.piece2slot[pc1] = sl2
        self.piece2slot[pc2] = sl1
        self.slot2piece[sl1] = pc2
        self.slot2piece[sl2] = pc1
        m = self.check_matches()

        if not m:
            # oops, invalid move, go back!
            self.piece2slot[pc1] = sl1
            self.piece2slot[pc2] = sl2
            self.slot2piece[sl1] = pc1
            self.slot2piece[sl2] = pc2
        else:
            self.logger.debug("switch piece from %s to %s",
                              sl1, sl2)
        return m


class PieceManager(cocos.layer.Layer):
    """Manager class for the gems."""

    def __init__(self, board, st, layout=None, crazy=False):
        super(PieceManager, self).__init__()
        self.logger = logging.getLogger('PieceManager')
        self.logger.info("Game start")
        self.board = board
        self.game_stats = st
        self.ph = PieceHolder()
        self.falling = False
        self.swapping = False
        self.ended = False
        self.paused = False
        self.repmatch = RepeatedMatcher(self)

        if layout is None:
            init_scenario = logic_board.generate_nonmatch_board()
        else:
            init_scenario = logic_board.generate_custom(layout)

        gems.crazy = crazy

        for x in range(8):
            for y in range(8):
                which = init_scenario[x][y]
                if which <= 6:
                    # normal pieces
                    gem = gems[which]
                elif which <= 12:
                    # explosives
                    gem = gems[which - 7].get_explosive()
                else:
                    # magic ball
                    gem = gems.get_magic_ball()
                self.game_stats.add_piece(gem)

                real_cords = self.board.get_slot_center(x, y)
                gem.position = real_cords
                gem.orig_pos = real_cords
                self.ph.set_piece(gem, x, y)
                self.add(gem)
        self.game_stats.redraw_stats()

        # labels
        self.pause_label = Label(_("(game paused)"), board.center,
                                 font_name='Sansation',
                                 font_size=32, color=(30, 30, 30, 255),
                                 anchor_x='center', anchor_y='center')
        self.newlevel_label = Label(_("New level!"),
                                    board.center, font_name='Sansation',
                                    font_size=50, color=(30, 30, 30, 255),
                                    anchor_x='center', anchor_y='center')

    def change_level(self):
        """Change level in the game."""
        def disappear():
            """Fade out current gems."""
            change_time = 1
            for gem in self.ph.piece2slot:
                gem.do(actions.FadeOut(change_time))
            self.newlevel_label.opacity = 0
            self.newlevel_label.do(actions.FadeIn(change_time))
            self.add(self.newlevel_label)
            pyglet.clock.schedule_once(populate, change_time)

        def populate(_):
            """Populate again."""
            self.remove(self.newlevel_label)

            # remove the gems from the layer (not visible now)
            quant_magic = 0
            quant_explosives = 0
            for gem in self.ph.piece2slot:
                if gem.is_explosive:
                    quant_explosives += 1
                if gem.is_magic:
                    quant_magic += 1
                self.remove(gem)

            # get a new scenario and load it
            self.ph = PieceHolder()
            specials = (quant_explosives, quant_magic)
            init_scenario = logic_board.generate_nonmatch_board(specials)

            for x in range(8):
                for y in range(8):
                    which = init_scenario[x][y]
                    if which <= 6:
                        # normal pieces
                        gem = gems[which]
                    elif which <= 13:
                        # explosives
                        gem = gems[which - 7].get_explosive()
                    else:
                        # magic ball
                        gem = gems.get_magic_ball()
                    self.game_stats.add_piece(gem)

                    real_cords = self.board.get_slot_center(x, y)
                    gem.position = real_cords
                    gem.orig_pos = real_cords
                    self.ph.set_piece(gem, x, y)
                    self.add(gem)
            self.game_stats.redraw_stats()

        def check(_):
            """Release the level change when all settled down."""
            if self.swapping or self.falling:
                pyglet.clock.schedule_once(check, .5)
            else:
                disappear()
        check(None)

    def set_pause(self, now_paused):
        """Sets or unsets the pause."""
        self.paused = now_paused
        pieces = self.ph.piece2slot.keys()
        if now_paused:
            new_pieces_opacity = 0
            self.add(self.pause_label)
        else:
            new_pieces_opacity = 255
            self.remove(self.pause_label)
        for p in pieces:
            p.opacity = new_pieces_opacity

    def hand_piece_select(self, piece):
        """Select a gem."""
        piece.scale *= 1.2

    def hand_piece_move(self, piece, coords):
        """Move the gem."""
        piece.position = coords

    def hand_piece_reset(self, piece):
        """Reset the gem movement."""
        piece.position = piece.orig_pos
        piece.scale = gems.resize_factor

    def hand_piece_switch(self, from_piece, to_piece):
        """Switch two gems."""
        self.swapping = True

        # move the selected one
        from_piece.do(actions.MoveTo(to_piece.orig_pos, PIECE_TRIP_TIME))

        # shrink the new one, and move it
        to_piece.scale /= 1.2
        to_piece.do(actions.MoveTo(from_piece.orig_pos, PIECE_TRIP_TIME))

        # schedule finish
        pyglet.clock.schedule_once(self._end_move, PIECE_TRIP_TIME,
                                   from_piece, to_piece)

    def _end_move(self, _, pc1, pc2):
        """Finish the move."""
        # we'll use the user move after finishing with board filling
        self.swapping = False
        if self.falling:
            pyglet.clock.schedule_once(self._end_move, .2, pc1, pc2)
            return
        self.falling = True

        # swap and check match
        m = self.ph.try_swap(pc1, pc2)

        if m:
            # yes! confirm the move
            # restart scales and swap their positions...
            pc1.scale = gems.resize_factor
            pc2.scale = gems.resize_factor
            pc1.orig_pos, pc2.orig_pos = pc2.orig_pos, pc1.orig_pos

            # score! score! score!
            self._matches(m, from_user=True, pieces_moved=(pc1, pc2))
        else:
            # move back the pieces
            backmove_time = PIECE_TRIP_TIME * .7
            pc1.do(actions.MoveTo(pc1.orig_pos, backmove_time))
            pc2.do(actions.MoveTo(pc2.orig_pos, backmove_time))

            def end_backmove(_):
                """Finish the back move."""
                self.hand_piece_reset(pc1)
                self.hand_piece_reset(pc2)
                self.falling = False
            pyglet.clock.schedule_once(end_backmove, backmove_time)

    def magic_trigger(self, pc1, pc2):
        """A magic ball was chosen over another one."""
        sounds.play("magicdisap")

        # find out which pieces to kill
        if pc1.is_magic and pc2.is_magic:
            remove_effect = remove_effects.Disappear((pc1, pc2))
        elif pc1.is_magic:
            pieces = []
            for piece, slot in self.ph.get_all_similar(pc2):
                pieces.append(piece)
            remove_effect = remove_effects.Magic(pc1, pieces)
        else:
            pieces = []
            for piece, slot in self.ph.get_all_similar(pc1):
                pieces.append(piece)
            remove_effect = remove_effects.Magic(pc2, pieces)

        self._remove_with_effects([remove_effect], {})

    def _matches(self, matches, from_user=False, pieces_moved=None):
        """There's a match, remove/blow the pieces and fill."""
        self.repmatch.hit(len(matches))
        ph = self.ph
        to_replace = {}
        removes = []
        special_scores = {}

        for pack in matches:
            pieces = [ph.get_piece(*slot) for slot in pack]
            explosives = [p for p in pieces if p.is_explosive]
            if explosives:
                sounds.play("explosion")
                special_scores[pack] = 'explode'

                to_explode = set()
                for piece in explosives:
                    # explode the explodable
                    to_explode.add(piece)

                    # and all around it
                    slot = ph.get_slot(piece)
                    around_slots = logic_board.get_around(slot)
                    arndpieces = (ph.get_piece(*slot) for slot in around_slots)
                    to_explode.update(set(arndpieces))

                to_disapp = [p for p in pieces if p not in to_explode]
                ef = remove_effects.Explode(to_explode, to_disapp, from_user)
                removes.append(ef)
            else:
                # normal case
                sounds.play("pumba")

                # remove'em all!
                removes.append(remove_effects.Disappear(pieces, from_user))

                # check special constructions
                if pieces_moved is None:
                    slot = logic_board.get_center(pack)
                else:
                    hints = [self.ph.get_slot(x) for x in pieces_moved]
                    slot = logic_board.get_center(pack, hints=hints)
                old_piece = ph.get_piece(*slot)

                len_pack = len(pack)
                if len_pack == 4:
                    # creates an explosive ball
                    new_piece = old_piece.get_explosive()
                elif len_pack == 5:
                    # magic!
                    new_piece = gems.get_magic_ball()
                else:
                    new_piece = None

                # if a new piece, make a relation for replacement
                if new_piece:
                    new_piece.orig_pos = old_piece.position
                    new_piece.position = old_piece.position
                    to_replace[old_piece] = new_piece

        self._remove_with_effects(removes, to_replace)

    def _remove_with_effects(self, remove_effect, to_replace):
        """Apply effects to remove."""
        to_refill = set()
        removed = set()
        full_wait = 0.1

        def remove(_, score_function, pieces):
            """Remove the pieces and score.

            From the layer, add to refill, and keep stats.
            """
            for p in pieces:
                if p not in removed:
                    self.remove(p)
                    removed.add(p)
                    self.game_stats.destroy_piece(p)

            score_data = score_function()
            if score_data is not None:
                self.game_stats.do_score(*score_data)

        for effect in remove_effect:
            (pieces_now, wait_time, pieces_after) = effect.remove(self)

            # remove some pieces now
            remove(None, effect.score_before, pieces_now)
            to_refill.update(set(pieces_now))

            # remove some pieces in the future, if any
            if pieces_after:
                pyglet.clock.schedule_once(remove, wait_time,
                                           effect.score_after, pieces_after)

                # need to get this now, because will schedule *now*
                # _fill_board_back to the future with these pieces
                to_refill.update(set(pieces_after))

            full_wait = max(full_wait, wait_time)

        # fill the board back after some time
        pyglet.clock.schedule_once(self._fill_board_back, full_wait,
                                   to_refill, to_replace)

    def _fill_board_back(self, _, to_refill, to_replace=None):
        """Remove pieces and fill the board again."""
        if to_replace is None:
            to_replace = {}

        ph = self.ph
        slots_to_remove = [ph.get_slot(p) for p in to_refill]

        for x in range(8):
            empties = 0
            for y in range(8):
                p = ph.get_piece(x, y)
                if (x, y) in slots_to_remove:
                    ph.remove_piece(x, y)

                    if p in to_replace:
                        # replace the gem with a specific new one
                        newpiece = to_replace[p]
                        if self.paused:
                            newpiece.opacity = 0
                        ph.set_piece(newpiece, x, y)
                        self.game_stats.add_piece(newpiece)
                        self.add(newpiece)
                        self._let_it_fall(newpiece, empties)
                    else:
                        # normal case, just increase empties to keep count of
                        # the gems that are gone in the column
                        empties += 1
                else:
                    if empties:
                        self._let_it_fall(p, empties)

            # fill the column with random gems entering from above
            for i in range(empties):
                upper_slot_y = 8 + i
                coord_x, coord_y = self.board.get_slot_center(x, upper_slot_y)
                newpiece = gems[random.randint(0, 6)]
                if self.paused:
                    newpiece.opacity = 0
                newpiece.orig_pos = coord_x, coord_y
                newpiece.position = coord_x, coord_y
                ph.set_piece(newpiece, x, upper_slot_y)
                self.game_stats.add_piece(newpiece)
                self.add(newpiece)
                self._let_it_fall(newpiece, empties)

        pyglet.clock.schedule_once(self._after_user_move, FALL_TIME)

    def _let_it_fall(self, piece, steps):
        """Move gems as if falling."""
        coord_x, old_cy = piece.orig_pos
        new_cy = old_cy - ARISTA * steps
        move = actions.MoveTo((coord_x, new_cy), FALL_TIME)
        piece.do(actions.Accelerate(move, 3))
        piece.orig_pos = coord_x, new_cy
        pyglet.clock.schedule_once(self.ph.piece_fall, FALL_TIME, piece, steps)

    def _after_user_move(self, _):
        """Finish the user movement."""
        # does the random-filled board has matches?
        m = self.ph.check_matches()
        if m:
            self._matches(m, from_user=False)
        else:
            self.falling = False
            # and there's any chance in the future?
            future = self.ph.future_guess()
            if not future:
                self.game_over()

    def game_over(self):
        """Sound and evaporation."""
        self.logger.info("Game over")
        self.ended = True
        sounds.play("game_over")

        for x in range(8):
            for y in range(8):
                p = self.ph.get_piece(x, y)
                p.do(cocos.actions.FadeOut(1.5))

        self.do(cocos.actions.Rotate(360 * 2, 3) |
                cocos.actions.ScaleTo(10, 3))

        self.game_stats.save_score()

    @property
    def is_not_ready(self):
        return self.swapping or self.ended or self.paused or self.falling


class MouseManager(cocos.layer.Layer):
    """Manager class for the mouse."""

    is_event_handler = True

    def __init__(self, piece_manager, board):
        super(MouseManager, self).__init__()
        self.pm = piece_manager
        self.board = board
        self.holding = None
        self.dragging = False
        self.drag_from = None

    def on_mouse_drag(self, coord_x, coord_y, delta_x, delta_y, buttons, mods):
        """User is dragging the cursor."""
        if self.pm.is_not_ready:
            return

        # check if moved the mouse enough to consider it a drag
        if self.drag_from is not None:
            dfx, dfy = self.drag_from
            if abs(coord_x - dfx) < 1 or abs(coord_y - dfy) < 1:
                return

        # ok, dragging
        coord_x, coord_y = director.get_virtual_coordinates(coord_x, coord_y)
        self.drag_from = None
        self.dragging = True

        if self.holding is None:
            slot = self.board.get_slot_from_cords(coord_x, coord_y)
            if slot is None:
                # outside the board!
                return
            piece = self.pm.ph.get_piece(*slot)
            if piece is None:
                # trying to get a blowed up one
                return
            self.holding = (slot, piece)
            self.pm.hand_piece_select(piece)

        # only move the piece if not selected it stationary
        _, piece = self.holding
        self.pm.hand_piece_move(piece, (int(coord_x), int(coord_y)))

    def on_mouse_release(self, coord_x, coord_y, buttons, modifiers):
        """Use it as indicator only when dragging."""
        if self.dragging:
            self.on_user_click(coord_x, coord_y)

    def on_mouse_press(self, coord_x, coord_y, buttons, modifiers):
        """User pressed a mouse button."""
        self.drag_from = coord_x, coord_y
        self.on_user_click(coord_x, coord_y)

    def on_user_click(self, coord_x, coord_y):
        """General click handling."""
        if self.pm.is_not_ready:
            return
        coord_x, coord_y = director.get_virtual_coordinates(coord_x, coord_y)

        new_slot = self.board.get_slot_from_cords(coord_x, coord_y)
        if new_slot is None:
            if self.holding is not None:
                # release it if click outside the board
                (old_slot, old_piece) = self.holding
                self.pm.hand_piece_reset(old_piece)
                self.holding = None
            return

        new_sx, new_sy = new_slot
        new_piece = self.pm.ph.get_piece(new_sx, new_sy)
        if new_piece is None:
            # trying to get or get into a blowed up one
            return

        if self.holding is None:
            self.pm.hand_piece_select(new_piece)
            self.holding = (new_slot, new_piece)
            return

        (old_slot, old_piece) = self.holding
        old_sx, old_sy = old_slot
        if old_sx == new_sx and old_sy == new_sy:
            # we're in the same slot, unselect it
            self.pm.hand_piece_reset(old_piece)
            self.holding = None
            self.dragging = False
            return

        diff = abs(new_sx - old_sx) + abs(new_sy - old_sy)
        if diff > 1:
            # too far, reset old one...
            (old_slot, old_piece) = self.holding
            self.pm.hand_piece_reset(old_piece)
            self.holding = None

            # .. and hold a new one if the previous one wasn't dragged
            if not self.dragging:
                self.pm.hand_piece_select(new_piece)
                self.holding = (new_slot, new_piece)
            else:
                self.dragging = False

            return

        if old_piece.is_magic or new_piece.is_magic:
            # special case where magic happens
            self.pm.magic_trigger(old_piece, new_piece)
        else:
            # move the original one
            self.pm.hand_piece_switch(old_piece, new_piece)

        self.holding = None
        self.dragging = False


class KeyboardManager(cocos.layer.Layer):
    """Handles the keyboard."""

    is_event_handler = True

    def __init__(self, game_manager):
        self.gm = game_manager
        cocos.layer.Layer.__init__(self)

    def on_key_press(self, key, modifiers):
        if key == keyboard.ESCAPE:
            self.gm.key_escape()
            return True

        if key in (keyboard.PAUSE, keyboard.P):
            self.gm.key_pause()
            return True

#        print "Key Pressed:", key, modifiers


class GameTiming(object):
    """Keeps the time of the game."""
    def __init__(self):
        self._start = time.time()
        self._is_paused = False
        self._time_paused = None

    def pause(self):
        """Stops counting the game time."""
        if self._is_paused:
            raise RuntimeError("Tried to pause while paused.")
        self._time_paused = time.time()
        self._is_paused = True

    def resume(self):
        """Starts counting game time again."""
        if not self._is_paused:
            raise RuntimeError("Tried to resume while not paused.")
        elapsed = time.time() - self._time_paused
        self._start += elapsed
        self._is_paused = False

    @property
    def time(self):
        """Returns the elapsed time."""
        if self._is_paused:
            t = self._time_paused - self._start
        else:
            t = time.time() - self._start
        return t


class GameManager(object):
    """Class that handles the games."""

    def __init__(self):
        self.playing = None
        self.timing = None
        self.config = ConfigManager()
        set_options(self.config)

    def _get_custom(self, specification):
        """Process the specification."""
        layout = []
        style = None
        for lin in open(specification):
            lin = lin.strip()
            if not lin or lin[0] == '#':
                continue

            if lin.startswith("style"):
                style = lin.split(":")[1].strip()
                continue

            numbers = map(int, lin.split())
            if len(numbers) != 8:
                raise ValueError("Bad numbers quantity in custom: %r" % lin)
            layout.append(numbers)

        if len(layout) != 8:
            raise ValueError("Bad rows quantity in custom: %d" % len(layout))

        # rotate clockwise
        rotated = []
        for col in range(8):
            newrow = [layout[row][col] for row in range(7, -1, -1)]
            rotated.append(newrow)

        return rotated, style

    def start(self, direct_play, specification):
        """Start a game!"""
        if direct_play:
            if direct_play == 'custom':
                layout, style = self._get_custom(specification)
                if style is None:
                    style = 'classic'
                sc = self._play_game(style, layout=layout)
            else:
                m = getattr(self, 'play_' + direct_play, None)
                if m is None:
                    print _("No such game!")
                    exit()
                sc = m()
        else:
            sc = scene.Scene()
            mm = MainMenu(self)
            sc.add(mm, z=1)
        return sc

    def play_classic(self):
        """Classic game."""
        return self._play_game('classic')

    def play_rush(self):
        """Rushed game."""
        return self._play_game('rush')

    def play_crazy(self):
        """Crazy game."""
        return self._play_game('crazy', crazy=True)

    def play_pyar(self):
        """Crazy game."""
        return self._play_game('pyar')

    def _get_background_image(self):
        """Return a randomly chosen background image."""
        images_dir = os.path.join(BASEDIR, "images", "backgrounds", "*.jpg")
        fname = os.path.basename(random.choice(glob.glob(images_dir)))
        return fname

    def _play_game(self, style, layout=None, crazy=False):
        """Play any game style."""
        self.timing = GameTiming()
        background_image = self._get_background_image()
        sc = scene.Scene(BackgroundLayer(background_image))
        board = Board()
        sc.add(board, z=0)

        # start stats and piece manager, but first set pieces style
        pieces = 'pyar' if style == 'pyar' else None
        gems.pieces = pieces
        st = stats.Stats(self, style)
        self.pm = PieceManager(board, st, layout=layout,
                               crazy=crazy)
        sc.add(self.pm, z=1)
        sc.add(MouseManager(self.pm, board), z=2)
        sc.add(KeyboardManager(self), z=2)
        sc.add(st, z=3)

        # add the image above with fade out, for a nice effect
        disap = BackgroundLayer(background_image)
        sc.add(disap, z=10)
        disap.do(actions.FadeOut(3) + actions.CallFunc(sc.remove, disap))

        self.playing = sc
        return sc

    @property
    def game_time(self):
        return self.timing.time

    def continue_game(self):
        """Keep playing the game."""
        if not self.pm.paused:
            self.timing.resume()
        director.replace(self.playing)

    def key_escape(self):
        """Go back to main menu, but keeping the game."""
        if not self.pm.paused:
            self.timing.pause()
        if self.pm.ended:
            self.playing = None
        self.main_menu()

    def key_pause(self):
        """Pause the game."""
        now_paused = not self.pm.paused
        if now_paused:
            self.timing.pause()
        else:
            self.timing.resume()
        self.pm.set_pause(now_paused)

    def main_menu(self):
        sc = scene.Scene()
        mm = MainMenu(self, self.playing)
        sc.add(mm, z=1)
        director.replace(sc)
