#!/usr/bin/env python

# Copyright 2011-2012 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/enjuewemela

"""Replay a game from a log."""

import operator
import re
import time

import pyglet

from pyglet.window import key as keyboard

import games
import utils

from cocos import scene, layer, actions
from cocos.director import director
from gems import gems


RE_ADD = ("(\d+-\d+-\d+ \d+:\d+:\d+),(\d+)  PieceHolder        "
          "DEBUG    set piece: ([\w-]+) in \((\d+), (\d+)\)\n")
RE_DEL = ("(\d+-\d+-\d+ \d+:\d+:\d+),(\d+)  PieceHolder        "
          "DEBUG    remove piece from \((\d+), (\d+)\): ([\w-]+)\n")
RE_SWITCH = ("(\d+-\d+-\d+ \d+:\d+:\d+),(\d+)  PieceHolder        DEBUG    "
             "switch piece from \((\d+), (\d+)\) to \((\d+), (\d+)\)\n")
RE_START = ("(\d+-\d+-\d+ \d+:\d+:\d+),(\d+)  PieceManager       INFO     "
            "Game start\n")

FH_FORMAT = "%Y-%m-%d %H:%M:%S"


class Operation(object):
    def __init__(self, **kwargs):
        keys = 'name tstamp piece x y xf yf xt yt'.split()
        self.__dict__.update(dict.fromkeys(keys))
        self.__dict__.update(kwargs)

    def __str__(self):
        data = ", ".join("%s=%s" % x for x in sorted(self.__dict__.items()))
        return "<Operation: %s>" % (data,)


def _file_parser(logfilename):
    """Parse log and yield data."""
    loginfo = []
    for line in open(logfilename):
        m = re.match(RE_START, line)
        if m:
            # game start, reset all and start again
            loginfo = []
            continue

        m = re.match(RE_ADD, line)
        if m:
            fh, ms, piece, x, y = m.groups()
            tstamp = time.mktime(time.strptime(fh, FH_FORMAT))
            tstamp += int(ms) / 1000.0
            x, y = int(x), int(y)
            o = Operation(name='add', tstamp=tstamp, piece=piece, x=x, y=y)
            loginfo.append(o)
            continue

        m = re.match(RE_DEL, line)
        if m:
            fh, ms, x, y, piece = m.groups()
            tstamp = time.mktime(time.strptime(fh, FH_FORMAT))
            tstamp += int(ms) / 1000.0
            x, y = int(x), int(y)
            o = Operation(name='del', tstamp=tstamp, piece=piece, x=x, y=y)
            loginfo.append(o)
            continue

        m = re.match(RE_SWITCH, line)
        if m:
            fh, ms, xf, yf, xt, yt = m.groups()
            tstamp = time.mktime(time.strptime(fh, FH_FORMAT))
            tstamp += int(ms) / 1000.0
            xf, yf, xt, yt = int(xf), int(yf), int(xt), int(yt)
            o = Operation(name='switch', tstamp=tstamp,
                          xf=xf, yf=yf, xt=xt, yt=yt)
            loginfo.append(o)
            continue
    return loginfo


class LogInfo(object):
    """Handles the log info."""
    def __init__(self, logfilename):
        self.loginfo = _file_parser(logfilename)
        self._pointer = None

    def next(self):
        """Return next log info."""
        if self._pointer is None:
            self._pointer = 0
        else:
            self._pointer += 1
        if self._pointer >= len(self.loginfo):
            return None
        return self.loginfo[self._pointer]

    def rewind(self):
        """Go back in one."""
        if self._pointer > 0:
            self._pointer -= 1


class Replayer(layer.Layer):
    """The replayer."""

    is_event_handler = True

    def __init__(self, board, loginfo):
        super(Replayer, self).__init__()
        self.board = board
        self.loginfo = loginfo
        self.placed = {}
        self.pending_falls = []
        self.history = []
        self.histptr = None

        # font and writer
        v_writer = utils.FontWriter(self, "Sansation")
        f_writer = utils.FontWriter(self, "SV Basic Manual")
        win_w, win_h = director.get_window_size()
        self._msg_label = v_writer("", (win_w // 2, 28), 16, "center",
                                   color=(0, 0, 255, 255))
        self.hist_label = f_writer("", (win_w - 25, 25), 11, "right",
                                   color=(0, 255, 0, 255))

    def show_message(self, text):
        """Show a message and hide it."""
        self._msg_label.element.text = text
        def _clean(_):
            """Clean the message."""
            self._msg_label.element.text = ""
        pyglet.clock.schedule_once(_clean, 1)

    def _fall(self, to_fill):
        """Make a piece fall."""
        # get other pieces in the column
        to_move = []
        pos_x, pos_y = to_fill
        while True:
            pos_y += 1
            piece = self.placed.get((pos_x, pos_y))
            if piece is None:
                break
            to_move.append(piece)

        # we have the list of pieces to move, from down up, let's move them
        pos_x, pos_y = to_fill
        for piece in to_move:
            gem = self.placed.pop((pos_x, pos_y + 1))
            new_coords = self.board.get_slot_center(pos_x, pos_y)
            gem.position = new_coords
            self.placed[(pos_x, pos_y)] = gem
            pos_y += 1

    def _rush(self):
        """Show changes that happened almost at the same time."""
        if self.pending_falls:
            to_fill = self.pending_falls.pop()
            self._fall(to_fill)
            return

        prvtstamp = None
        while True:
            op = self.loginfo.next()
            if op is None:
                if self.pending_falls:
                    break
                else:
                    # all is really done, end
                    self.show_message("Hit END")
                    return True

            # detect a time lapse, and put a fake pause in it
            if prvtstamp is None:
                prvtstamp = op.tstamp
            if op.tstamp - prvtstamp > .2:
                self.loginfo.rewind()
                break

            if op.name == 'add':
                real_coords = self.board.get_slot_center(op.x, op.y)
                gem = gems.by_name(op.piece)
                gem.position = real_coords
                self.board.add(gem)
                self.placed[(op.x, op.y)] = gem
            elif op.name == 'del':
                gem = self.placed.pop((op.x, op.y))
                gem.do(
                    actions.FadeOut(.1) + actions.FadeIn(.1) +
                    actions.FadeOut(.1) + actions.FadeIn(.1) +
                    actions.CallFunc(self.board.remove, gem)
                )
                self.pending_falls.append((op.x, op.y))
            elif op.name == 'switch':
                gemf = self.placed[(op.xf, op.yf)]
                gemt = self.placed[(op.xt, op.yt)]
                self.placed[(op.xf, op.yf)] = gemt
                self.placed[(op.xt, op.yt)] = gemf

                def _setpos(gem, position):
                    """Set the position."""
                    gem.position = position

                gemf.do(
                    actions.FadeOut(.1) + actions.FadeIn(.1) +
                    actions.FadeOut(.1) + actions.FadeIn(.1) +
                    actions.CallFunc(_setpos, gemf, gemt.position)
                )
                gemt.do(
                    actions.FadeOut(.1) + actions.FadeIn(.1) +
                    actions.FadeOut(.1) + actions.FadeIn(.1) +
                    actions.CallFunc(_setpos, gemt, gemf.position)
                )
                break

            prvtstamp = op.tstamp

        # reorder pending falls to let them attack in order, and assure
        # all are valid (maybe it was re-filled with an explosive or something)
        self.pending_falls.sort(key=operator.itemgetter(1))
        self.pending_falls = [x for x in self.pending_falls
                              if x not in self.placed]

    def _copy_placed(self):
        """Copy the placed dict, but using gem name and not object."""
        d = {}
        for pos, gem in self.placed.iteritems():
            d[pos] = str(gem)
        return d

    def start(self):
        """Go."""
        self.histptr = 0
        self.hist_label.element.text = '0'
        self._rush()
        new_placed = self._copy_placed()
        self.history.append(new_placed)

    def _go_forward(self):
        """Move forward."""
        if self.histptr + 1 == len(self.history):
            hit_end = self._rush()
            if not hit_end:
                self.histptr += 1
                new_placed = self._copy_placed()
                self.history.append(new_placed)
        else:
            self.histptr += 1
            self._reshow(self.history[self.histptr])
        self.hist_label.element.text = str(self.histptr)

    def _go_back(self):
        """Move back."""
        if self.histptr is None or self.histptr == 0:
            self.show_message("Hit BEGIN")
            return
        self.histptr -= 1
        self.hist_label.element.text = str(self.histptr)
        self._reshow(self.history[self.histptr])

    def _reshow(self, new_image):
        """Clean and show a scenario from scratch."""
        for gem in self.placed.itervalues():
            self.board.remove(gem)
        self.placed = {}
        for (x, y), gem_name in new_image.items():
            real_coords = self.board.get_slot_center(x, y)
            gem = gems.by_name(gem_name)
            gem.position = real_coords
            self.board.add(gem)
            self.placed[(x, y)] = gem

    def on_key_press(self, key, modifiers):
        """Receive a key press event."""
        if key in (keyboard.SPACE, keyboard.RIGHT, keyboard.DOWN):
            self._go_forward()
        elif key in (keyboard.LEFT, keyboard.UP):
            self._go_back()


def go(logfilename):
    """Main entry point."""
    loginfo = LogInfo(logfilename)

    sc = scene.Scene()
    board = games.Board()
    sc.add(board, z=0)

    r = Replayer(board, loginfo)
    sc.add(r, z=0)
    r.start()

    return sc
